static char sccsid[] =
    "%Z%%P%: %I% %E% %U%\n"
/*
########################################################################
#*MODULE HEADER*
#
# Program  : dtfn (%M%)
# Version  : %I%
# Date/Time: %G% %U%
# For      : The World
#
#-----------------------------------------------------------------------
#*DESCRIPTION*
#
#   cc -o dtfn dtfn.c
#
#   Valid range of dates is from 1753 to 2799
#
#   Julian period is a chronological system used by astronomers and is
#   based on the consecutive numbering of days from Jan. 1, 4713 BC.
#   Not to be confused with the Julian calendar, the Julian period was
#   devised by the scholar Joseph Justus Scaliger in 1582 and named by
#   him for his father, Julius Caesar Scaliger. Joseph Scaliger proposed
#   a period of 7,980 years of numbered days to be used in determining
#   time elapsed between various historical events otherwise recorded
#   only in different chronologies, eras or calendars. The length of
#   7,980 years was choosen as the product of 28 x 19 x 15; these
#   respectively, are the number of years in the so called solar cycle
#   of the Julian calendar in which dates recur on the same days of the
#   week; the lunar or Metonic cycle, after which phases of the Moon
#   recur on a particular day in the solar year, or year of seasons;
#   and the cycle of indiction, originally a schedule of period taxes
#   or government requisitions in ancient Rome.  The Epoch, or starting
#   point of 4713 BC was chosen as the nearest past year in which the
#   three cycles began together. January 1st, 2000 is JD 2,451,545.
#
#   Leap years are those divisible by 4 except centesimal years
#   (years ending with 00) which are common unless divisible by 400.
#
#   A Century consists of 100 consecutive calendar years.  The 1st
#   century consisted of the years 1 through 100.  The 20th century
#   consisted of the years 1901 through 2000.  The 21st century will
#   begin on January 1st, 2001.
#
#      September 1752
#    S  M Tu  W Th  F  S
#          1  2 14 15 16
#   17 18 19 20 21 22 23
#   24 25 26 27 28 29 30
#
#   1582 Oct 5 - Oct 10
#   1752 Sep 2 - Sep 14
#   1918 Feb 1 - Feb 14
#
#   Daylight Saving Time switchover spring date for USA
#            (First Sunday of April 01:59:59 -> 03:00:00)
#   Daylight Saving Time switchover fall date for USA
#            (Last Sunday of October 02:00:00 -> 01:00:01)
#
#   Change from 2007:
#   Daylight Saving Time switchover spring date for USA
#            (Second Sunday of March 01:59:59 -> 03:00:00)
#   Daylight Saving Time switchover fall date for USA
#            (First Sunday of November 02:00:00 -> 01:00:01)
#
#   UNIX = 2,147,483,647  1/18/2038 22:14:07 EST
#        = 2,147,498,047  1/19/2038 03:14:07 UTC
#
#   This program is designed to acquire a date "relative" to a given
#   date. You will give it a number and the program will add or
#   subtract that number from the julian day number of the string
#   date and will return the new string date value. The program can
#   compare two dates and calculate the number of days between them.
#   It can verify the entered date and return 9 if it is not valid
#   or format it to return the date in the standard format.
#   It can convert between Oracle (DD-MON-YYYY) and standard date
#   format (MM/DD/YYYY).
#   It can supply different date formats for file extensions, etc.
#-----------------------------------------------------------------------
#
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>
#include <getopt.h>  /*for Linux */

struct DATE_INFO
{
    int  yyyy;
    int  mm;
    int  dd;
    long jd;
    int  wday;
    int  yday;
    int  leap;
};

struct TIME_INFO
{
    int  hh;
    int  mi;
    int  ss;
};

int  datevals_YMD(char *s, struct DATE_INFO *pd);
int  datevals_DMY(char *s, struct DATE_INFO *pd);
int  datevals_MDY(char *s, struct DATE_INFO *pd);
int  datevals2(char *s, struct DATE_INFO *pd);
int  timevals(char *s, struct TIME_INFO *pd);
int  days_in_month(int year, int month);
int  leap_year(int year);
long days_between_dates(struct DATE_INFO *pd, struct DATE_INFO *pd2);
int  day_of_year(struct DATE_INFO *pd);
long day_of_century(struct DATE_INFO *pd);
int  day_of_week(long julian);
long date_to_julian(struct DATE_INFO *pd);
void julian_to_date(struct DATE_INFO *pd);
int  adjust_century(int year);
int  adjust_year(struct DATE_INFO *pd);
int  yyjjj(struct DATE_INFO *pd);
void g_easter(struct DATE_INFO *pd);
void upperCase(char *s);

double JD_date_part(double jdd);
double JD_time_part(double jdd);
double date_and_time_to_JDD(struct DATE_INFO *d, struct TIME_INFO *t);
double time_to_fraction(struct TIME_INFO *t);
struct TIME_INFO fraction_to_time(double frac);
long long_mod (long n, long modulus);

static int day_tab [2][13] = {
    {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
    {0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}
};

static char day_name[7][4] =
    {"Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"};

static char day_name2[7][10] =
    {"Monday", "Tuesday", "Wednesday", "Thursday", "Friday",
     "Saturday", "Sunday"};

static char mm_name[13][4] =
    {"NON", "JAN", "FEB", "MAR", "APR", "MAY", "JUN",
     "JUL", "AUG", "SEP", "OCT", "NOV", "DEC"};

static char mm_naml[13][4] =
    {"Bug", "Jan", "Feb", "Mar", "Apr", "May", "Jun",
     "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

static char mm_name_full[13][10] =
    {"Bug", "January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December"};

static char *usage_str="\nUsage: %s [-options] Otto P. Bedrosian  V1.20  2006-11-10\n\n"
"Default date format is mm/dd/[yy]yy (US) unless specified otherwise.\n\n"
"If date is not entered, system date is substituted if appropriate.\n\n"
"If 2-digit year is entered it will be expanded to 4-digit using\n"
"a rolling window based on system date +10 and -90 years.\n\n"
"Valid range of dates is from 1753 to 2799.\n\n"
"${T1} is date and time in the format 'mm/dd/yyyy hh:mi:ss'.\n\n"
"-A months [mm/dd/yyyy]     Adjust the date by the number of months.\n"
"-a days   [mm/dd/yyyy]     Adjust the date by the number of days.\n"
"-B [mm/dd/yyyy]            Beginning of the week (Sunday) for a date.\n"
"-b [mm/dd/yyyy]            First workday of the month for a date.\n"
"-C [mm/dd/yyyy]            Day of century for the date.\n"
"-c mm/dd/yyyy [mm/dd/yyyy] Compare date1 and date2. Return (date1 - date2).\n"
"-D [mm/dd/yyyy]            Name of the day for a date (Monday).\n"
"-d [mm/dd/yyyy]            Abbreviated name of the day for a date (Mon).\n"
"-E [mm/dd/yyyy]            End of week (Saturday) for a date.\n"
"-e [yyyy]                  Date of Easter for a year.\n"
"-F days [mm/dd/yyyy]       Adjust the date by the number of workdays.\n"
"-f mm/dd/yyyy [mm/dd/yyyy] Workdays between date1 and date2 (inclusive).\n"
"-G minutes [${T1}]         Adjust the date/time by minutes ('yyyy-mm-dd hh:mi:ss').\n"
"-g minutes [${T1}]         Adjust the date/time by minutes ('mm/dd/yyyy hh:mi:ss').\n"
"-H [yyyy-mm-dd]            Abbreviated name of the day for a date (Mon).\n"
"-I [yyyy]                  Daylight Saving Time spring switchover date for US.\n"
"-i [yyyy]                  Daylight Saving Time fall switchover date for US.\n"
"-J JD number               Date from Julian Date (2,361,222 - 2,743,738).\n"
"-j [mm/dd/yyyy]            Julian Date number for the date (1753 - 2799).\n"
"-K [mm/dd/yyyy]            Number of days in a year 'yyyyddd'.\n"
"-k [yy][yy]ddd             Date from number of days in a year 'mm/dd/yyyy'.\n"
"-L [mm/dd/yyyy]            Last day of the month for a date.\n"
"-l [mm/dd/yyyy]            Last workday of the month for a date.\n"
"-M [mm/dd/yyyy]            Name of the month for the date (January).\n"
"-m [mm/dd/yyyy]            Abbreviated name of the month for the date (Jan).\n"
"-N [mm/dd/yyyy]            Work hours in the month for a date (8 hrs/day).\n"
"-n weeks [mm/dd/yyyy]      Adjust the date by the number of weeks.\n"
"-O [dd-MON-yyyy]           Convert date from 'dd-MON-yyyy' to 'mm/dd/yyyy'.\n"
"-o [mm/dd/yyyy]            Convert date from 'mm/dd/yyyy' to 'dd-MON-yyyy'.\n"
"-P hours [mm/dd/yyyy]      Adjust the date by work hours (8 hrs/day).\n"
"-p mm/dd/yyyy [mm/dd/yyyy] Work hours between date1 and date2 (8 hrs/day).\n"
"-Q [yyyy-mm-dd]            Quarter from date (yyyy-mm-dd to yyyy_qq).\n"
"-q [mm/dd/yyyy]            Quarter from date (mm/dd/yyyy to yyyy_qq).\n"
"-R ${T1} [${T2}]           Elapsed time in seconds (input format 'yyyy-mm-dd').\n"
"-r ${T1} [${T2}]           Elapsed time in 'hh:mi:ss' (input 'yyyy-mm-dd').\n"
"-S                         System date and time (yyyy-mm-dd hh:mi:ss).\n"
"-s                         System date and time (yyyy-mm-dd hh:mi:ss Day).\n"
"-T                         System date and time (mm/dd/yyyy hh:mi:ss).\n"
"-t                         System date and time (mm/dd/yyyy hh:mi:ss Day).\n"
"-U ${T1} [${T2}]           Elapsed time between two events in seconds.\n"
"-u ${T1} [${T2}]           Elapsed time between two events in 'hh:mi:ss'.\n"
"-V yyyy-mm-dd              Verify date: if ok return (yyyy-mm-dd) else '9'.\n"
"-v mm/dd/yyyy              Verify date: if ok return (mm/dd/yyyy) else '9'.\n"
"-W [yyyy-mm-dd]            Day of week (0-6, Sunday=0, Monday=1, Saturday=6).\n"
"-w [mm/dd/yyyy]            Day of week (0-6, Sunday=0, Monday=1, Saturday=6).\n"
"-X [mm/dd/yyyy]            Date in format 'yyyymmdd_doy_w_Day'.\n"
"-x                         System date and time in 'yyyymmdd_hhmiss_w_Day'.\n"
"-Y [yyyy-mm-dd]            Convert date from format yyyy-mm-dd to mm/dd/yyyy.\n"
"-y [mm/dd/yyyy]            Convert date from format mm/dd/yyyy to yyyy-mm-dd.\n"
"-Z [dd-mm-yyyy]            Convert date from format dd-mm-yyyy to mm/dd/yyyy.\n"
"-z [mm/dd/yyyy]            Convert date from format mm/dd/yyyy to dd-mm-yyyy.\n"
"-? or -h                   List of options. (%s -? > dtfn.txt)\n";

static char *usage_str2="\nUsage: %s -?  Get option list (help) V1.20 2006-11-10\n";

static char *usage_str3="\nUsage: %s -u date1 time1 [date2] [time2]\n\n"
"  Returns the elapsed time between two events.\n\n"
"  date1     first date value  mm/dd/[yy]yy.\n"
"  time1     first time value  hh:mi[:ss].\n\n"
"  date2     second date value mm/dd/[yy]yy.\n"
"  time2     second time value hh:mi[:ss].\n\n"
"  Example:  T1=`date +'%%D %%T'`  or  T1=`date +'%%m-%%d-%%Y %%T'`\n"
"            ELAPSED=`%s -u ${T1} ${T2}`\n\n"
"  If the second date and time value is not supplied\n"
"  then the current system date and time is used.\n\n";

static char *usage_str4="\nUsage: %s -g minutes [date] [time]\n\n"
"  Returns the date and time as a result of the added minutes.\n\n"
"  minutes   minutes to be added or subtracted.\n\n"
"  date      date value  mm/dd/[yy]yy.\n"
"  time      time value  hh:mi[:ss].\n\n"
"  Example:  T1=`date +'%%D %%T'`  or  T1=`date +'%%m-%%d-%%Y %%T'`\n"
"            RESULT=`%s -g minutes ${T1}`\n\n"
"  If the date and time value is not supplied\n"
"  then the current system date and time is used.\n\n";

static char *usage_str5="\nUsage: %s -r date1 time1 [date2] [time2]\n\n"
"  Returns the elapsed time between two events.\n\n"
"  date1     first date value  [yy]yy-mm-dd.\n"
"  time1     first time value  hh:mi[:ss].\n\n"
"  date2     second date value [yy]yy-mm-dd.\n"
"  time2     second time value hh:mi[:ss].\n\n"
"  Example:  T1=`date +'%%Y-%%m-%%d %%T'`\n"
"            ELAPSED=`%s -r ${T1} ${T2}`\n\n"
"  If the second date and time value is not supplied\n"
"  then the current system date and time is used.\n\n";

static char *usage_str6="\nUsage: %s -G minutes [date] [time]\n\n"
"  Adjust the date and time by the number of minutes.\n\n"
"  date      date value  [yy]yy-mm-dd.\n"
"  time      time value  hh:mi[:ss].\n\n"
"  Example:  T1=`date +'%%Y-%%m-%%d %%T'`\n"
"            NEW_TIME=`%s -G minutes ${T1}`\n\n"
"  If the date and time value is not supplied\n"
"  then the current system date and time is used.\n\n";

static char *optstring ="A:a:BbCcDdEeF:fG:g:HhIiJ:jKk:LlMmNn:OoP:pQqRrSsTtUuVvWwXxYyZz?";

/***************************************************************************/

int main(int argc, char *argv[])
{
    struct DATE_INFO dat0, dat1, dat2;
    struct TIME_INFO tim0, tim1, tim2;
    struct tm *tinfo;
    time_t run_time;
    char   date_str0[11], date_str1[11], date_str2[11];
    char   time_str0[9],  time_str1[9],  time_str2[9];

    int  option;           /* use with getopt */
    char output_date[11];  /* string value of output date */
    long add_to_date;      /* option value to adjust date by */
    long date_diff;        /* difference of two dates */
    long jul_date0, jul_date1, jul_date2;
    double jdd, jdd1, jdd2;
    int  doy;              /* day of year */
    int  vdd, vmm;
    int  no_date;          /* flag to signal no date requirement */
    int  two_dates;        /* flag to signal two date requirement */
    int  todays;           /* flag to signal today's date */
    char action_switch[2]; /* action variable for switch stmt */

    long days1, days2, seconds1, seconds2, elapsed, l_time, l_days;
    int  i_hour, i_mins, i_secs;
    char output_str[25];

    add_to_date = no_date = two_dates = todays = 0;

    setbuf(stdout, (char *) 0);        /* for debugging purposes */

    /*-----------------------------------------------------------------
     |  Get current system date                                       |
     *----------------------------------------------------------------*/

    (void) time(&run_time);
    tinfo = localtime(&run_time);

    sprintf(date_str0, "%.4d-%.2d-%.2d",
            tinfo->tm_year + 1900, tinfo->tm_mon + 1, tinfo->tm_mday);

    if (datevals_YMD(date_str0, &dat0))
    {
        fprintf(stderr, "\nInvalid system date. Program aborted.\n");
        exit(1);
    }
    dat1.jd = dat0.jd = jul_date0 = date_to_julian(&dat0);
    dat0.wday = day_of_week(dat0.jd);
    julian_to_date(&dat1);

    sprintf(date_str1, "%.2d/%.2d/%.4d", dat1.mm, dat1.dd, dat1.yyyy);

    sprintf(time_str0, "%.2d:%.2d:%.2d",
            tinfo->tm_hour, tinfo->tm_min, tinfo->tm_sec);

    if (timevals(time_str0, &tim0))
    {
        fprintf(stderr, "\nInvalid system time. Program aborted.\n");
        exit(1);
    }

    /*-----------------------------------------------------------------
     |  getopt                                                        |
     *----------------------------------------------------------------*/

    while((option = getopt(argc, argv, optstring)) != EOF)
    {
    switch (option)
        {
        case 'A':
            action_switch[0]='A';
            action_switch[1]='\0';
            add_to_date = atol(optarg);
            break;
        case 'a':
            action_switch[0]='a';
            action_switch[1]='\0';
            add_to_date = atol(optarg);
            break;
        case 'b':
            action_switch[0]='b';
            action_switch[1]='\0';
            break;
        case 'C':
            action_switch[0]='C';
            action_switch[1]='\0';
            break;
        case 'c':
            two_dates = 1;
            action_switch[0]='c';
            action_switch[1]='\0';
            break;
        case 'D':
            action_switch[0]='D';
            action_switch[1]='\0';
            break;
        case 'd':
            action_switch[0]='d';
            action_switch[1]='\0';
            break;
        case 'B':
            action_switch[0]='B';
            action_switch[1]='\0';
            break;
        case 'E':
            action_switch[0]='E';
            action_switch[1]='\0';
            break;
        case 'e':
            no_date = 1;
            action_switch[0]='e';
            action_switch[1]='\0';
            break;
        case 'F':
            action_switch[0]='F';
            action_switch[1]='\0';
            add_to_date = atol(optarg);
            break;
        case 'f':
            two_dates = 1;
            action_switch[0]='f';
            action_switch[1]='\0';
            break;
        case 'G':
            no_date = 1;
            action_switch[0]='G';
            action_switch[1]='\0';
            add_to_date = atol(optarg);
            break;
        case 'g':
            action_switch[0]='g';
            action_switch[1]='\0';
            add_to_date = atol(optarg);
            break;
        case 'H':
            no_date = 1;
            action_switch[0]='H';
            action_switch[1]='\0';
            break;
        case 'I':
            no_date = 1;
            action_switch[0]='I';
            action_switch[1]='\0';
            break;
        case 'i':
            no_date = 1;
            action_switch[0]='i';
            action_switch[1]='\0';
            break;
        case 'J':
            no_date = 1;
            action_switch[0]='J';
            action_switch[1]='\0';
            jul_date1 = atol(optarg);
            if (jul_date1 < 2361222 || jul_date1 > 2743738)
            {
                fprintf(stderr, usage_str2, basename(argv[0]));
                exit(1);
            }
            dat1.jd = jul_date1;
            julian_to_date(&dat1);
            break;
        case 'j':
            action_switch[0]='j';
            action_switch[1]='\0';
            break;
        case 'K':
            action_switch[0]='K';
            action_switch[1]='\0';
            break;
        case 'k':
            no_date = 1;
            action_switch[0]='k';
            action_switch[1]='\0';
            jul_date1 = atol(optarg);
            if (jul_date1 < 1 || jul_date1 > 2799365)
            {
                fprintf(stderr, usage_str2, basename(argv[0]));
                exit(1);
            }
            if (jul_date1 < 367)
            {
                jul_date1 += ((long) dat0.yyyy * 1000);
            }
            if (jul_date1 > 1000 && jul_date1 < 99366)
            {
                doy = (int) (jul_date1 / 1000);
                doy = adjust_century(doy);
                jul_date1 = jul_date1 % 1000;
                jul_date1 += ((long) doy * 1000);
            }
            dat1.jd = jul_date1;
            if (yyjjj(&dat1))
            {
                fprintf(stderr, usage_str2, basename(argv[0]));
                exit(1);
            }
            break;
        case 'L':
            action_switch[0]='L';
            action_switch[1]='\0';
            break;
        case 'l':
            action_switch[0]='l';
            action_switch[1]='\0';
            break;
        case 'M':
            action_switch[0]='M';
            action_switch[1]='\0';
            break;
        case 'm':
            action_switch[0]='m';
            action_switch[1]='\0';
            break;
        case 'N':
            action_switch[0]='N';
            action_switch[1]='\0';
            break;
        case 'n':
            action_switch[0]='n';
            action_switch[1]='\0';
            add_to_date = atol(optarg);
            break;
        case 'o':
            action_switch[0]='o';
            action_switch[1]='\0';
            break;
        case 'O':
            no_date = 1;
            action_switch[0]='O';
            action_switch[1]='\0';
            break;
        case 'P':
            action_switch[0]='P';
            action_switch[1]='\0';
            add_to_date = atol(optarg);
            break;
        case 'p':
            two_dates = 1;
            action_switch[0]='p';
            action_switch[1]='\0';
            break;
        case 'Q':
            no_date = 1;
            action_switch[0]='Q';
            action_switch[1]='\0';
            break;
        case 'q':
            action_switch[0]='q';
            action_switch[1]='\0';
            break;
        case 'R':
            no_date = 1;
            action_switch[0]='R';
            action_switch[1]='\0';
            break;
        case 'r':
            no_date = 1;
            action_switch[0]='r';
            action_switch[1]='\0';
            break;
        case 'S':
            sprintf(output_date, "%.4d-%.2d-%.2d %.2d:%.2d:%.2d",
                    dat0.yyyy, dat0.mm, dat0.dd,
                    tim0.hh, tim0.mi, tim0.ss);
            fprintf(stdout, "%s\n", output_date);
            exit(0);
            break;
        case 's':
            sprintf(output_date, "%.4d-%.2d-%.2d %.2d:%.2d:%.2d %s",
                    dat0.yyyy, dat0.mm, dat0.dd,
                    tim0.hh, tim0.mi, tim0.ss,
                    day_name[dat0.wday]);
            fprintf(stdout, "%s\n", output_date);
            exit(0);
            break;
        case 'T':
            sprintf(output_date, "%.2d/%.2d/%.4d %.2d:%.2d:%.2d",
                    dat0.mm, dat0.dd, dat0.yyyy,
                    tim0.hh, tim0.mi, tim0.ss);
            fprintf(stdout, "%s\n", output_date);
            exit(0);
            break;
        case 't':
            sprintf(output_date, "%.2d/%.2d/%.4d %.2d:%.2d:%.2d %s",
                    dat0.mm, dat0.dd, dat0.yyyy,
                    tim0.hh, tim0.mi, tim0.ss,
                    day_name[dat0.wday]);
            fprintf(stdout, "%s\n", output_date);
            exit(0);
            break;
        case 'U':
            no_date = 1;
            action_switch[0]='U';
            action_switch[1]='\0';
            break;
        case 'u':
            no_date = 1;
            action_switch[0]='u';
            action_switch[1]='\0';
            break;
        case 'V':
            action_switch[0]='V';
            action_switch[1]='\0';
            break;
        case 'v':
            action_switch[0]='v';
            action_switch[1]='\0';
            break;
        case 'W':
            no_date = 1;
            action_switch[0]='W';
            action_switch[1]='\0';
            break;
        case 'w':
            action_switch[0]='w';
            action_switch[1]='\0';
            break;
        case 'X':
            action_switch[0]='X';
            action_switch[1]='\0';
            break;
        case 'x':
            action_switch[0]='x';
            action_switch[1]='\0';
            break;
        case 'Y':
            no_date = 1;
            action_switch[0]='Y';
            action_switch[1]='\0';
            break;
        case 'y':
            action_switch[0]='y';
            action_switch[1]='\0';
            break;
        case 'Z':
            no_date = 1;
            action_switch[0]='Z';
            action_switch[1]='\0';
            break;
        case 'z':
            action_switch[0]='z';
            action_switch[1]='\0';
            break;
        case '?':
        case 'h':
            fprintf(stdout, usage_str, basename(argv[0]), basename(argv[0]));
            exit(0);
            break;
        }
    }

    /*-----------------------------------------------------------------
     | Check for the right amount of variables on the command line.   |
     *----------------------------------------------------------------*/

    if (argc == 1)
    {
        fprintf(stdout, usage_str2, basename(argv[0]));
        exit(1);
    }

    if ((no_date == 1) && (argc - optind) < 1)
    {
        todays = 1;
/*      fprintf(stderr, usage_str2, basename(argv[0]));
        exit(1); */
    }

    if ( two_dates && (argc - optind) < 2 )
    {
        todays = 1;
/*      fprintf(stderr, usage_str2, basename(argv[0]));
        exit(1); */
    }

    if ( two_dates && (argc - optind) < 1 )
    {
        fprintf(stderr, usage_str2, basename(argv[0]));
        exit(1);
    }

    /*-----------------------------------------------------------------
     |  Convert the string argument into a julian number for          |
     |  starting and ending date.                                     |
     *----------------------------------------------------------------*/

    if ((argc - optind) > 0) strcpy(date_str1, argv[optind]);

    if ( no_date )
    {
        if (action_switch[0] == 'U'
         || action_switch[0] == 'u'
         || action_switch[0] == 'R'
         || action_switch[0] == 'r')
        {
          if (action_switch[0] == 'R'
           || action_switch[0] == 'r') {
          sprintf(date_str2, "%.4d-%.2d-%.2d",
                  tinfo->tm_year + 1900, tinfo->tm_mon + 1, tinfo->tm_mday);
          } else {
          sprintf(date_str2, "%.2d/%.2d/%.4d",
                  tinfo->tm_mon + 1, tinfo->tm_mday, tinfo->tm_year + 1900);
          }
          sprintf(time_str2, "%.2d:%.2d:%.2d",
                  tinfo->tm_hour, tinfo->tm_min, tinfo->tm_sec);
          if ( argc < 4 ) {
            fprintf(stdout, usage_str5, argv[0], argv[0]);
            exit(1);
          }
          strcpy(date_str1, argv[2]);
          strcpy(time_str1, argv[3]);
          if ( argc == 5 ) {
            strcpy(date_str2, argv[4]);
            strcpy(time_str2, "00:00");
          }
          if ( argc == 6 ) {
            strcpy(date_str2, argv[4]);
            strcpy(time_str2, argv[5]);
          }
          if (action_switch[0] == 'R'
           || action_switch[0] == 'r') {
          if (datevals_YMD(date_str1, &dat1)) {
          fprintf(stderr, "\nInvalid first date argument. Program aborted.\n");
          fprintf(stderr, usage_str5, argv[0], argv[0]);
          exit(1);
          }
          } else {
          if (datevals_MDY(date_str1, &dat1)) {
          fprintf(stderr, "\nInvalid first date argument. Program aborted.\n");
          fprintf(stderr, usage_str3, argv[0], argv[0]);
          exit(1);
          }
          }
          if (timevals(time_str1, &tim1)) {
          fprintf(stderr, "\nInvalid first time argument.  Program aborted.\n");
          fprintf(stderr, usage_str3, argv[0], argv[0]);
          exit(1);
          }
          if (action_switch[0] == 'R'
           || action_switch[0] == 'r') {
          if (datevals_YMD(date_str2, &dat2)) {
          fprintf(stderr, "\nInvalid first date argument. Program aborted.\n");
          fprintf(stderr, usage_str5, argv[0], argv[0]);
          exit(1);
          }
          } else {
          if (datevals_MDY(date_str2, &dat2)) {
          fprintf(stderr, "\nInvalid second date argument. Program aborted.\n");
          fprintf(stderr, usage_str3, argv[0], argv[0]);
          exit(1);
          }
          }
          if (timevals(time_str2, &tim2)) {
          fprintf(stderr, "\nInvalid second time argument. Program aborted.\n");
          fprintf(stderr, usage_str3, argv[0], argv[0]);
          exit(1);
          }
          days1 = date_to_julian(&dat1);
          seconds1 = (long) (tim1.hh * 3600 + tim1.mi * 60 + tim1.ss);

          days2 = date_to_julian(&dat2);
          seconds2 = (long) (tim2.hh * 3600 + tim2.mi * 60 + tim2.ss);

          if (labs(days1 - days2) < 24855)
          {
            seconds1 += (days1 * 86400);
            seconds2 += (days2 * 86400);
          }
          else
          {
            fprintf(stderr, "\nOut of range.  Program aborted.\n");
            fprintf(stderr, usage_str3, argv[0], argv[0]);
            exit(1);
          }

          elapsed = seconds2 - seconds1;
          if (action_switch[0] == 'R'
           || action_switch[0] == 'U')
          {
            elapsed = seconds2 - seconds1;
            sprintf(output_str, "%ld", elapsed);
            fprintf(stdout, "%s\n", output_str);
          }
          else
          {
          elapsed = labs(seconds1 - seconds2);

          l_days = elapsed / 86400;
          l_time = elapsed % 86400;
          i_secs = (int) (l_time % 60);
          l_time = l_time / 60;
          i_mins = (int) (l_time % 60);
          i_hour = (int) (l_time / 60);

          if ( l_days >= 1 && l_days < 4) i_hour += (int) (l_days * 24);

          if ( l_days < 4 ) {
            sprintf(time_str2, "%.2d:%.2d:%.2d", i_hour, i_mins, i_secs);
            fprintf(stdout, "%s\n", time_str2);
          }
          else if ( l_days > 2 || l_days < 1000) {
            sprintf(output_str, "%3ld days and %.2d:%.2d:%.2d",
                l_days, i_hour, i_mins, i_secs);
            fprintf(stdout, "%s\n", output_str);
          }
          else
          {
            sprintf(output_str, "%5ld days and %.2d:%.2d:%.2d",
                l_days, i_hour, i_mins, i_secs);
            fprintf(stdout, "%s\n", output_str);
          }
          }
          exit(0);
        }
/*                                                         */
        if (action_switch[0] == 'O')
        {
          if (argc == 3) {
            if (datevals2(date_str1, &dat1))
            {
                fprintf(stderr, "\nInvalid input date [dd-MON-yyyy].\n");
                fprintf(stderr, usage_str2, basename(argv[0]));
                exit(1);
            }
          }
        }
        else if (action_switch[0] == 'Q'
              || action_switch[0] == 'Y'
              || action_switch[0] == 'H'
              || action_switch[0] == 'W')
        {
          if (argc == 3) {
            if (datevals_YMD(date_str1, &dat1))
            {
                fprintf(stderr, "\nInvalid input date [yyyy-mm-dd].\n");
                fprintf(stderr, usage_str2, basename(argv[0]));
                exit(1);
            }
          }
        }
        else if (action_switch[0] == 'Z')
        {
          if (argc == 3) {
            if (datevals_DMY(date_str1, &dat1))
            {
                fprintf(stderr, "\nInvalid input date [dd-mm-yyyy].\n");
                fprintf(stderr, usage_str2, basename(argv[0]));
                exit(1);
            }
          }
        }
        else if (action_switch[0] == 'e' ||
                 action_switch[0] == 'I' ||
                 action_switch[0] == 'i')
        {
            if ((argc - optind) < 1)
            {
                dat1.yday = dat0.yyyy;
            }
            else
            {
                jul_date1 = atoi(date_str1);
                if (jul_date1 < 1 || jul_date1 > 2799)
                {
                    fprintf(stderr, usage_str2, basename(argv[0]));
                    exit(1);
                }
                if (jul_date1 > 99 && jul_date1 < 1753)
                {
                    fprintf(stderr, usage_str2, basename(argv[0]));
                    exit(1);
                }
                dat1.yday = jul_date1;
            }
            if (action_switch[0] == 'e') g_easter(&dat1);
            if (action_switch[0] == 'I')
            {
                dat1.yyyy = dat1.yday;
                dat1.mm = 4;
                dat1.dd = 1;
                if (dat1.yyyy > 2006)
                {
                   dat1.mm = 3;
                   dat1.dd = 8;
                }
                dat1.jd = date_to_julian(&dat1);
                while (day_of_week(dat1.jd) != 6) dat1.jd++;
                julian_to_date(&dat1);
                dat1.wday = day_of_week(dat1.jd);
            }
            if (action_switch[0] == 'i')
            {
                dat1.yyyy = dat1.yday;
                dat1.mm = 10;
                dat1.dd = 31;
                if (dat1.yyyy > 2006)
                {
                   dat1.mm = 11;
                   dat1.dd = 1;
                   dat1.jd = date_to_julian(&dat1);
                   while (day_of_week(dat1.jd) != 6) dat1.jd++;
                   julian_to_date(&dat1);
                   dat1.wday = day_of_week(dat1.jd);
                }
                else
                {
                dat1.jd = date_to_julian(&dat1);
                while (day_of_week(dat1.jd) != 6) dat1.jd--;
                julian_to_date(&dat1);
                dat1.wday = day_of_week(dat1.jd);
                }
            }
        }
    }
    else
    {
        if (action_switch[0] == 'v' ||
            action_switch[0] == 'V')
        {
            if ((argc - optind) < 1)
            {
                fprintf(stdout, "9\n");
                exit(1);
            }
            if (action_switch[0] == 'v')
              no_date = datevals_MDY(date_str1, &dat1);
            else
              no_date = datevals_YMD(date_str1, &dat1);
            if (no_date)
            {
    /*          fprintf(stdout, "%d\n", no_date);  */
                fprintf(stdout, "9\n");
                exit(1);
            }
        }
        else
        {
            if (! todays)
            {
                if (datevals_MDY(date_str1, &dat1))
                {
                    fprintf(stderr, "\nInvalid input date. Program aborted.\n");
                    fprintf(stderr, usage_str2, basename(argv[0]));
                    exit(1);
                }
            }
            else
            {
                dat1.jd = jul_date0;
                julian_to_date(&dat1);
            }
            jul_date1 = date_to_julian(&dat1);
        }
    }

    if ( two_dates )
    {
        if (! todays)
        {
            strcpy(date_str2, argv[optind + 1]);
            if (datevals_MDY(date_str2, &dat2))
            {
                fprintf(stderr, "\nInvalid input date. Program aborted.\n");
                fprintf(stderr, usage_str2, basename(argv[0]));
                exit(1);
            }
            jul_date2 = date_to_julian(&dat2);
        }
        else
        {
            strcpy(date_str2, argv[optind]);
            if (datevals_MDY(date_str2, &dat2))
            {
                fprintf(stderr, "\nInvalid input date. Program aborted.\n");
                fprintf(stderr, usage_str2, basename(argv[0]));
                exit(1);
            }
            jul_date2 = date_to_julian(&dat2);
        }
    }

    switch ( action_switch[0] )
    {
    case 'A':
        dat2.jd = jul_date1;
        julian_to_date(&dat2);
        dat2.dd = 1;
        vdd = (add_to_date < 0) ? -1 : 1;
        vmm = (int) abs(add_to_date);
        while (vmm != 0)
        {
            dat2.mm += vdd;
            if (dat2.mm < 1)
            {
               dat2.yyyy--;
               dat2.mm = 12;
            }
            if (dat2.mm > 12)
            {
               dat2.yyyy++;
               dat2.mm = 1;
            }
            vmm--;
        }
        dat2.dd = dat1.dd;
        if (days_in_month(dat2.yyyy, dat2.mm) < dat2.dd)
        {
          dat2.dd = days_in_month(dat2.yyyy, dat2.mm);
        }
        if (dat2.yyyy < 1753 || dat2.yyyy > 2799)
        {
            fprintf(stderr, "\nValid date between 1753 and 2799.\n");
        }
        else
        {
            sprintf(output_date, "%.2d/%.2d/%.4d", dat2.mm, dat2.dd, dat2.yyyy);
            fprintf(stdout, "%s\n", output_date);
        }
        break;
    case 'a':
        dat2.jd = jul_date1 + add_to_date;
        if (dat2.jd < 2361222 || dat2.jd > 2743738)
        {
          fprintf(stderr, "\nResult outside valid date range of 1753 and 2799.\n");
        }
        else
        {
          julian_to_date(&dat2);
          sprintf(output_date, "%.2d/%.2d/%.4d", dat2.mm, dat2.dd, dat2.yyyy);
          fprintf(stdout, "%s\n", output_date);
        }
        break;
    case 'B':
        dat1.jd = date_to_julian(&dat1);
        while (day_of_week(dat1.jd) != 6) dat1.jd--;
        julian_to_date(&dat1);
        dat1.wday = day_of_week(dat1.jd);
        sprintf(output_date, "%.2d/%.2d/%.4d", dat1.mm, dat1.dd, dat1.yyyy);
        fprintf(stdout, "%s\n", output_date);
        break;
    case 'b':
        dat1.dd = 1;
        dat1.jd = date_to_julian(&dat1);
        while (day_of_week(dat1.jd) > 4) dat1.jd++;
        julian_to_date(&dat1);
        dat1.wday = day_of_week(dat1.jd);
        sprintf(output_date, "%.2d/%.2d/%.4d", dat1.mm, dat1.dd, dat1.yyyy);
        fprintf(stdout, "%s\n", output_date);
        break;
    case 'C':
        date_diff = day_of_century(&dat1);
        fprintf(stdout, "%ld\n", date_diff);
        break;
    case 'c':
        date_diff = jul_date1 - jul_date2;
        fprintf(stdout, "%ld\n", date_diff);
        break;
    case 'D':
        dat1.jd = date_to_julian(&dat1);
        dat1.wday = day_of_week(dat1.jd);
        fprintf(stdout, "%s\n", day_name2[dat1.wday]);
        break;
    case 'H':
    case 'd':
        dat1.jd = date_to_julian(&dat1);
        dat1.wday = day_of_week(dat1.jd);
        fprintf(stdout, "%s\n", day_name[dat1.wday]);
        break;
    case 'E':
        dat1.jd = date_to_julian(&dat1);
        while (day_of_week(dat1.jd) != 5) dat1.jd++;
        julian_to_date(&dat1);
        dat1.wday = day_of_week(dat1.jd);
        sprintf(output_date, "%.2d/%.2d/%.4d", dat1.mm, dat1.dd, dat1.yyyy);
        fprintf(stdout, "%s\n", output_date);
        break;
    case 'G':
        sprintf(date_str1, "%.4d-%.2d-%.2d",
                tinfo->tm_year + 1900, tinfo->tm_mon + 1, tinfo->tm_mday);
        sprintf(time_str1, "%.2d:%.2d:%.2d",
                tinfo->tm_hour, tinfo->tm_min, tinfo->tm_sec);
        if ( argc == 4 ) {
          strcpy(date_str1, argv[3]);
          strcpy(time_str1, "00:00");
        }
        if ( argc == 5 ) {
          strcpy(date_str1, argv[3]);
          strcpy(time_str1, argv[4]);
        }
        if (datevals_YMD(date_str1, &dat1)) {
          fprintf(stderr, "\nInvalid input date. Program aborted.\n");
          fprintf(stderr, usage_str6, argv[0], argv[0]);
          exit(1);
        }
        if (timevals(time_str1, &tim1)) {
          fprintf(stderr, "\nInvalid input time. Program aborted.\n");
          fprintf(stderr, usage_str6, argv[0], argv[0]);
          exit(1);
        }
        jdd = date_and_time_to_JDD(&dat1, &tim1);
        jdd1 = (double) add_to_date;
        jdd1 /= 1440.0;
        jdd += jdd1;
        dat2.jd = (long) JD_date_part(jdd);
        jdd = JD_time_part(jdd);
        julian_to_date(&dat2);
        tim2 = fraction_to_time(jdd);

        sprintf(date_str2, "%.4d-%.2d-%.2d", dat2.yyyy, dat2.mm, dat2.dd);
        sprintf(time_str2, "%.2d:%.2d:%.2d", tim2.hh, tim2.mi, tim2.ss);
        fprintf(stdout, "%s %s\n", date_str2, time_str2);
        break;
    case 'g':
        sprintf(date_str1, "%.2d/%.2d/%.4d",
                tinfo->tm_mon + 1, tinfo->tm_mday, tinfo->tm_year + 1900);
        sprintf(time_str1, "%.2d:%.2d:%.2d",
                tinfo->tm_hour, tinfo->tm_min, tinfo->tm_sec);
        if ( argc == 4 ) {
          strcpy(date_str1, argv[3]);
          strcpy(time_str1, "00:00");
        }
        if ( argc == 5 ) {
          strcpy(date_str1, argv[3]);
          strcpy(time_str1, argv[4]);
        }
        if (datevals_MDY(date_str1, &dat1)) {
          fprintf(stderr, "\nInvalid input date. Program aborted.\n");
          fprintf(stderr, usage_str4, argv[0], argv[0]);
          exit(1);
        }
        if (timevals(time_str1, &tim1)) {
          fprintf(stderr, "\nInvalid input time. Program aborted.\n");
          fprintf(stderr, usage_str4, argv[0], argv[0]);
          exit(1);
        }
        jdd = date_and_time_to_JDD(&dat1, &tim1);
        jdd1 = (double) add_to_date;
        jdd1 /= 1440.0;
        jdd += jdd1;
        dat2.jd = (long) JD_date_part(jdd);
        jdd = JD_time_part(jdd);
        julian_to_date(&dat2);
        tim2 = fraction_to_time(jdd);

        sprintf(date_str2, "%.2d/%.2d/%.4d", dat2.mm, dat2.dd, dat2.yyyy);
        sprintf(time_str2, "%.2d:%.2d:%.2d", tim2.hh, tim2.mi, tim2.ss);
        fprintf(stdout, "%s %s\n", date_str2, time_str2);
        break;
    case 'j':
        jul_date1 = date_to_julian(&dat1);
        fprintf(stdout, "%ld\n", jul_date1);
        break;
    case 'K':
        doy = day_of_year(&dat1);
        fprintf(stdout, "%.4d%.3d\n", dat1.yyyy, doy);
        break;
    case 'L':
        dat1.dd = days_in_month(dat1.yyyy, dat1.mm);
        dat1.jd = date_to_julian(&dat1);
        julian_to_date(&dat1);
        sprintf(output_date, "%.2d/%.2d/%.4d", dat1.mm, dat1.dd, dat1.yyyy);
        fprintf(stdout, "%s\n", output_date);
        break;
    case 'l':
        dat1.mm++;
        dat1.dd = 1;
        if (dat1.mm > 12)
        {
            dat1.mm = 1;
            dat1.yyyy++;
        }
        dat1.jd = date_to_julian(&dat1);
        do
        {
            dat1.jd--;
        } while (day_of_week(dat1.jd) > 4);
        julian_to_date(&dat1);
        dat1.wday = day_of_week(dat1.jd);
        sprintf(output_date, "%.2d/%.2d/%.4d", dat1.mm, dat1.dd, dat1.yyyy);
        fprintf(stdout, "%s\n", output_date);
        break;
    case 'M':
        fprintf(stdout, "%s\n", mm_name_full[dat1.mm]);
        break;
    case 'm':
        fprintf(stdout, "%s\n", mm_naml[dat1.mm]);
        break;
    case 'e':
    case 'J':
    case 'k':
    case 'I':
    case 'i':
    case 'O':
    case 'Y':
    case 'v':
    case 'Z':
        sprintf(output_date, "%.2d/%.2d/%.4d", dat1.mm, dat1.dd, dat1.yyyy);
        fprintf(stdout, "%s\n", output_date);
        break;
    case 'o':
        sprintf(output_date, "%.2d-%s-%.4d",
                dat1.dd, mm_name[dat1.mm], dat1.yyyy);
        fprintf(stdout, "%s\n", output_date);
        break;
    case 'F':
    case 'P':
        dat2.jd = jul_date1;
        if (add_to_date != 0)
        {
          vdd = (add_to_date < 0) ? -1 : 1;
          add_to_date = labs(add_to_date);
          if (action_switch[0] == 'P') add_to_date = (long) ((add_to_date + 7) / 8);
          add_to_date--;
          while (add_to_date != 0)
          {
            if (vdd == 1)
              dat2.jd++;
            else
              dat2.jd--;
            if(day_of_week(dat2.jd) < 5) add_to_date--;
          }
        }
        julian_to_date(&dat2);
        sprintf(output_date, "%.2d/%.2d/%.4d", dat2.mm, dat2.dd, dat2.yyyy);
        fprintf(stdout, "%s\n", output_date);
        break;
    case 'f':
    case 'p':
        jul_date1 = dat1.jd = date_to_julian(&dat1);
        jul_date2 = dat2.jd = date_to_julian(&dat2);
        if (jul_date2 < jul_date1) {
           days1 = jul_date2;
           jul_date2 = jul_date1;
           jul_date1 = days1;
           dat1.jd = jul_date1;
           dat2.jd = jul_date2;
        }
        /* Date of the previous Monday for the starting date */
        while (day_of_week(jul_date1) != 0) jul_date1--;
        /* Date of the previous Monday for the ending date */
        while (day_of_week(jul_date2) != 0) jul_date2--;
        days2 = (dat2.jd - jul_date2) - 4;
        days1 = (dat1.jd - jul_date1) - 5;
        days2 = (days2 < 0) ? 0 : days2;
        days1 = (days1 < 0) ? 0 : days1;
        /* Take the difference between the dates and add 1 (inclusive) */
        date_diff = (dat2.jd - dat1.jd) + 1;
        /* Subtract the number of weekends */
        date_diff -= (((jul_date2 - jul_date1) / 7) * 2) + days2 - days1;
        if (action_switch[0] == 'p') date_diff *= 8; /* Multiply by 8 hours per day */
        fprintf(stdout, "%ld\n", date_diff);
        break;
    case 'N':
        dat1.dd = 1;
        jul_date1 = dat2.jd = dat1.jd = date_to_julian(&dat1);
        julian_to_date(&dat2);
        dat2.dd = days_in_month(dat2.yyyy, dat2.mm);
        jul_date2 = dat2.jd = date_to_julian(&dat2);
        /* Date of the previous Monday for the starting date */
        while (day_of_week(jul_date1) != 0) jul_date1--;
        /* Date of the previous Monday for the ending date */
        while (day_of_week(jul_date2) != 0) jul_date2--;
        days2 = (dat2.jd - jul_date2) - 4;
        days1 = (dat1.jd - jul_date1) - 5;
        days2 = (days2 < 0) ? 0 : days2;
        days1 = (days1 < 0) ? 0 : days1;
        /* Take the absolute difference between the dates and add 1 */
        date_diff = labs(dat2.jd - dat1.jd) + 1;
        /* Subtract the number of weekends */
        date_diff -= ((labs(jul_date2 - jul_date1) / 7) * 2) + days2 - days1;
        /* Multiply by 8 hours per day */
        date_diff *= 8;
        fprintf(stdout, "%ld\n", date_diff);
        break;
    case 'n':
        dat2.jd = jul_date1 + (add_to_date * 7);
        julian_to_date(&dat2);
        sprintf(output_date, "%.2d/%.2d/%.4d", dat2.mm, dat2.dd, dat2.yyyy);
        fprintf(stdout, "%s\n", output_date);
        break;
    case 'Q':
    case 'q':
        i_hour = (int) (((dat1.mm - 1) / 3) + 1);
        sprintf(output_date, "%.4d_%.2d", dat1.yyyy, i_hour);
        fprintf(stdout, "%s\n", output_date);
        break;
    case 'W':
    case 'w':
        dat1.jd = date_to_julian(&dat1);
        dat1.wday = day_of_week(dat1.jd);
        doy = dat1.wday + 1;
        if (doy == 7) doy = 0;
        fprintf(stdout, "%d\n", doy);
        break;
    case 'V':
    case 'y':
        sprintf(output_date, "%.4d-%.2d-%.2d", dat1.yyyy, dat1.mm, dat1.dd);
        fprintf(stdout, "%s\n", output_date);
        break;
    case 'X':
        dat1.jd = date_to_julian(&dat1);
        dat1.wday = day_of_week(dat1.jd);
        doy = dat1.wday + 1;
        if (doy == 7) doy = 0;
        vdd = day_of_year(&dat1);
        sprintf(output_date, "%.4d%.2d%.2d_%.3d_%.1d_%s",
                dat1.yyyy, dat1.mm, dat1.dd,
                vdd, doy, day_name[dat1.wday]);
        fprintf(stdout, "%s\n", output_date);
        break;
    case 'x':
        doy = dat0.wday + 1;
        if (doy == 7) doy = 0;
        sprintf(output_date, "%.4d%.2d%.2d_%.2d%.2d%.2d_%.1d_%s",
                dat0.yyyy, dat0.mm, dat0.dd,
                tim0.hh, tim0.mi, tim0.ss,
                doy, day_name[dat0.wday]);
        fprintf(stdout, "%s\n", output_date);
        break;
    case 'z':
        sprintf(output_date, "%.2d-%.2d-%.4d", dat1.dd, dat1.mm, dat1.yyyy);
        fprintf(stdout, "%s\n", output_date);
        break;
    case '?':
        fprintf(stderr, usage_str, argv[0], argv[0], argv[0]);
        exit(1);
        break;
    }
    return(0);
}

/*----------------------------------------------------------------------*
 |    end main
 *----------------------------------------------------------------------*/
/***************************************************************************/

int datevals_YMD (char *s, struct DATE_INFO *pd)
{
    char *ptr, temp[5];
    int  i, vals[3];

    ptr = s;
    i = 0;

    for (;;)
    {
        if (isdigit(*s))
        {
            s++;
            if (*s != '\0')
            {
                continue;
            }
        }
        strncpy(temp, ptr, s - ptr);
        temp[s - ptr] = '\0';

        vals[i++] = atoi(temp);
        if (*s == '\0')
        {
            break;
        }
        s++;
        ptr = s;
    }

    if (vals[1] < 1 || vals[1] > 12) return(1);

    if (vals[0] < 100) vals[0] = adjust_century(vals[0]);

    if (vals[0] < 1753 || vals[0] > 2799) return(2);

    if (vals[2] < 1 || vals[2] > days_in_month(vals[0], vals[1])) return(3);

    pd->yyyy = vals[0];
    pd->mm = vals[1];
    pd->dd = vals[2];

    return(0);
}

int datevals_DMY (char *s, struct DATE_INFO *pd)
{
    char *ptr, temp[5];
    int  i, vals[3];

    ptr = s;
    i = 0;

    for (;;)
    {
        if (isdigit(*s))
        {
            s++;
            if (*s != '\0')
            {
                continue;
            }
        }
        strncpy(temp, ptr, s - ptr);
        temp[s - ptr] = '\0';

        vals[i++] = atoi(temp);
        if (*s == '\0')
        {
            break;
        }
        s++;
        ptr = s;
    }

    if (vals[1] < 1 || vals[1] > 12) return(1);

    if (vals[2] < 100) vals[2] = adjust_century(vals[2]);

    if (vals[2] < 1753 || vals[2] > 2799) return(2);

    if (vals[0] < 1 || vals[0] > days_in_month(vals[2], vals[1])) return(3);

    pd->dd = vals[0];
    pd->mm = vals[1];
    pd->yyyy = vals[2];

    return(0);
}

int datevals_MDY (char *s, struct DATE_INFO *pd)
{
    char *ptr, temp[5];
    int  i, vals[3];

    ptr = s;
    i = 0;

    for (;;)
    {
        if (isdigit(*s))
        {
            s++;
            if (*s != '\0')
            {
                continue;
            }
        }
        strncpy(temp, ptr, s - ptr);
        temp[s - ptr] = '\0';

        vals[i++] = atoi(temp);
        if (*s == '\0')
        {
            break;
        }
        s++;
        ptr = s;
    }

    if (vals[0] < 1 || vals[0] > 12) return(1);

    if (vals[2] < 100) vals[2] = adjust_century(vals[2]);

    if (vals[2] < 1753 || vals[2] > 2799) return(2);

    if (vals[1] < 1 || vals[1] > days_in_month(vals[2], vals[0])) return(3);

    pd->mm = vals[0];
    pd->dd = vals[1];
    pd->yyyy = vals[2];

    return(0);
}

int datevals2 (char *s, struct DATE_INFO *pd)
{
    char *ptr, temp[5], mmm[4];
    int  i, vals[3];

    sscanf(s, "%d-%3s-%d", &vals[1], mmm, &vals[2]);

    upperCase(mmm);

    for ( i=1; i<13; i++)
    {
        if (! strcmp(mm_name[i], mmm))
        {
            vals[0] = i;
            break;
        }
    }

    if (vals[0] < 1 || vals[0] > 12) return(1);

    if (vals[2] < 100) vals[2] = adjust_century(vals[2]);

    if (vals[2] < 1753 || vals[2] > 2799) return(2);

    if (vals[1] < 1 || vals[1] > days_in_month(vals[2], vals[0])) return(3);

    pd->mm = vals[0];
    pd->dd = vals[1];
    pd->yyyy = vals[2];

    return(0);
}

int timevals (char *s, struct TIME_INFO *pd)
{
    char *ptr, temp[5];
    int  i, vals[3];
    vals[0] = 0;
    vals[1] = 0;
    vals[2] = 0;

    ptr = s;
    i = 0;

    for (;;)
    {
        if (isdigit(*s))
        {
            s++;
            if (*s != '\0')
            {
                continue;
            }
        }
        strncpy(temp, ptr, s - ptr);
        temp[s - ptr] = '\0';

        vals[i++] = atoi(temp);
        if (*s == '\0')
        {
            break;
        }
        s++;
        ptr = s;
    }

    if (vals[0] < 0 || vals[0] > 23) return(1);

    if (vals[1] < 0 || vals[1] > 59) return(2);

    if (vals[2] < 0 || vals[2] > 59) vals[2] = 0;

    pd->hh = vals[0];
    pd->mi = vals[1];
    pd->ss = vals[2];

    return(0);
}

int days_in_month(int year, int month)
{
    int days, leap;

    leap = leap_year(year);
    days = day_tab[leap][month];
    return(days);
}

int leap_year(int year)
{
    if ((year % 4 == 0) && (year % 100 != 0) || (year % 400 == 0))
        return(1);
    else
        return(0);
}

long days_between_dates(struct DATE_INFO *pd, struct DATE_INFO *pd2)
{
    long days;

    days = labs(date_to_julian(pd) - date_to_julian(pd2));
    return(days);
}

int day_of_year(struct DATE_INFO *pd)
{
    struct DATE_INFO dat_tmp;
    dat_tmp.yyyy = pd->yyyy - 1;
    dat_tmp.mm = 12;
    dat_tmp.dd = 31;

    return( (int) (days_between_dates(pd, &dat_tmp)));
}

long day_of_century(struct DATE_INFO *pd)
{
    struct DATE_INFO dat_tmp;
    long   days;

    dat_tmp.yyyy = pd->yyyy - (pd->yyyy % 100);
    dat_tmp.mm = 12;
    dat_tmp.dd = 31;

    days = date_to_julian(pd) - date_to_julian(&dat_tmp);
    if (days < 1)
    {
        dat_tmp.yyyy -= 100;
        days = date_to_julian(pd) - date_to_julian(&dat_tmp);
    }
    return(days);
}

/*--FUNCTION    julian_to_date
 *
 *  Description:
 *      Calculates the date from the Julian Date number
 *
 *  Arguments:
 *      DATE_INFO structure with JD populated
 *
 *  Returns:
 *      DATE_INFO structure with yyyy, mm, dd
 *
 */
void julian_to_date(struct DATE_INFO *pd)
{
    long x, y, d, m;
    int  day, month, year;

    x = 4 * pd->jd - 6884477;
    y = (x / 146097) * 100;
    d = (x % 146097) / 4;

    x = 4 * d + 3;
    y = (x / 1461) + y;
    d = (x % 1461) / 4 + 1;

    x = 5 * d - 3;
    m = x / 153 + 1;
    d = ((x % 153) / 5) + 1;

    if (m < 11)
        month = m + 2;
    else
        month = m - 10;

    day = d;

    year = y + m / 11;

    pd->yyyy = year;
    pd->mm = month;
    pd->dd = day;
}

/*--FUNCTION    date_to_julian
 *
 *  Description:
 *      Calculates the Julian Date number from the date
 *
 *  Arguments:
 *      DATE_INFO structure with YYYY, MM, DD populated
 *
 *  Returns:
 *      DATE_INFO structure with JD
 *
 */
long date_to_julian(struct DATE_INFO *pd)
{
    int  year, month, day;
    long ta, tb, tc, julian;

    day = pd->dd;
    month = pd->mm;
    year = pd->yyyy;

    if (month > 2)
    {
        month = month - 3;
    }
    else
    {
        month = month + 9;
        year = year - 1;
    }
    ta = (146097 * (year / 100)) / 4;
    tb = (long) (1461 * ((float) (year % 100) / 4));
    tc = (153 * month + 2) / 5 + day + 1721119;

    julian = ta + tb + tc;
    return(julian);
}

/*--FUNCTION    day_of_week
 *
 *  Description:
 *      Function will calculate the day of the week
 *
 *  Arguments:
 *      A long integer Julian Date number
 *
 *  Returns:
 *      An integer from 0 to 6. Values correspond
 *      to days of the week as follows:
 *
 *      0 = Monday
 *      1 = Tuesday
 *      2 = Wednesday
 *      3 = Thursday
 *      4 = Friday
 *      5 = Saturday
 *      6 = Sunday
 */
int day_of_week(long julian)
{
    int index;
    index = (int) (julian % 7L);
/*
    index = (int) ((julian % 7L) + 1);
    if (index == 7) index = 0;
*/
    return(index);
}

/*--FUNCTION    adjust_century
 *
 *  Description:
 *      When only 2-digit year is given the function will calculate
 *      the 4-digit year by using a rolling window of +10 and -90
 *      years from the current system date
 *
 *  Arguments:
 *      An integer 2-digit year
 *
 *  Returns:
 *      An integer 4-digit year
 *
 */
int adjust_century(int year)
{
    struct  tm *tpd;
    time_t  run_time;
    int     cc, yy;

    (void) time(&run_time);
    tpd = localtime(&run_time);

    yy = tpd->tm_year + 1900;
    cc = yy / 100;
    yy = yy % 100;

    if (year > (yy + 10))
        year += (cc - 1) * 100;
    else
    {
        if (year < (yy - 89))
            year += (cc + 1) * 100;
        else
            year += cc * 100;
    }
    return(year);
}

int adjust_year(struct DATE_INFO *pd)
{
    struct tm *tpd;
    time_t run_time;
    struct DATE_INFO dat_tmp;
    int    max_days = 365;

    (void) time(&run_time);
    tpd = localtime(&run_time);

    pd->yyyy = tpd->tm_year + 1900;

    if (pd->yday > (tpd->tm_yday) + 1) pd->yyyy--;

    if (leap_year(pd->yyyy)) max_days = 366;

    if (pd->yday > max_days) return(1);

    dat_tmp.yyyy = pd->yyyy - 1;
    dat_tmp.mm = 12;
    dat_tmp.dd = 31;
    pd->jd = date_to_julian(&dat_tmp);
    pd->jd += pd->yday;
    julian_to_date(pd);

    return(0);
}

int yyjjj(struct DATE_INFO *pd)
{
    struct DATE_INFO   dat_tmp;
    int    max_days = 365;

    pd->yyyy = (int) (pd->jd / 1000);

    if (pd->yyyy < 100)
    pd->yyyy = adjust_century(pd->yyyy);

    if (pd->yyyy < 1753 || pd->yyyy > 2799) return(1);

    if (leap_year(pd->yyyy)) max_days = 366;

    pd->yday = (int) (pd->jd % 1000);

    if (pd->yday < 1 || pd->yday > max_days) return(1);

    dat_tmp.yyyy = pd->yyyy - 1;
    dat_tmp.mm = 12;
    dat_tmp.dd = 31;
    pd->jd = date_to_julian(&dat_tmp);
    pd->jd += pd->yday;
    julian_to_date(pd);

    return(0);
}

double JD_date_part(double jdd)
{
    double frac, jd0;
    long   jdl;

    jdl = (long) jdd;
    jd0 = (double) jdl;
    frac = jdd - jd0;

    if (frac >= 0.5) jd0 += 1.0;

    return(jd0);
}

double JD_time_part(double jdd)
{
    double frac, jd0;
    long   jdl;

    jdl = (long) jdd;
    jd0 = (double) jdl;
    frac = jdd - jd0;

    if (frac >= 0.5) frac -= 1.0;

    return(frac);
}

double date_and_time_to_JDD(struct DATE_INFO *d, struct TIME_INFO *t)
{
    return( date_to_julian(d) + time_to_fraction(t) );
}

double time_to_fraction(struct TIME_INFO *t)
{
    double frac, h, m, s;

    h = (double) (t->hh);
    m = (double) (t->mi);
    s = (double) (t->ss);

    h /= 24.0;
    m /= 1440.0;
    s /= 86400.0;

    frac = (h + m + s) - 0.5;

    return(frac);
}

struct TIME_INFO fraction_to_time(double frac)
{
    struct TIME_INFO t;
    double f, h, m, s;
    long hl, ml, sl;

    f = frac + 0.5;
    h = f * 24.0;
    hl = (long) h;
    f = h - ((double) hl);
    m = f * 60.0;
    ml = (long) m;
    f = m - ((double) ml);
    s = f * 60.0 + 0.5;
    sl = (long) s;

    t.hh = (int) hl;
    t.mi = (int) ml;
    t.ss = (int) sl;

    if (t.ss == 60)
    {
        t.ss = 0;
        t.mi++;
    }
    if (t.mi == 60)
    {
        t.mi = 0;
        t.hh++;
    }

    if (t.hh == 24) t.hh = 0;

    return(t);
}

void g_easter(struct DATE_INFO *pd)
{
    int  year;
    long y, g, c, x, z, d, e, n, e1, n1, dpn;

    year = pd->yday;

    if (year < 100) year = adjust_century(year);

    y = (long) year;
    g = (y - ((y / 19L) * 19L)) + 1L;
    c = (y / 100L) + 1L;
    x = ((3L * c) / 4L) - 12L;
    z = (((8L * c) + 5L) / 25L) - 5L;
    d = ((5L * y) / 4L) - x - 10L;
    e1 = ((11L * g) + 20L + z - x);
    e = e1 - ((e1 / 30L) * 30L);
    while (e < 0L) e += 30L;
    if ( ((e == 25L) && (g > 11)) || (e == 24L) ) e++;
    n = 44L - e;
    if (n < 21L) n += 30L;
    dpn = d + n;
    n1 = dpn - ((dpn / 7L) * 7L);
    n = n + 7L - n1;
    if (n > 31L)
    {
        pd->mm = 4;
        pd->dd = (int) (n - 31L);
    }
    else
    {
        pd->mm = 3;
        pd->dd = (int) n;
    }
    pd->yyyy = year;
}

long long_mod (long n, long modulus)
{
    long n1;

    n1 = n - (n / modulus) * modulus;

    return(n1);
}

void upperCase (char *pS)
{
    while (*pS)
    {
        if (islower (*pS))
        {
            *pS = toupper(*pS);
        }
        pS++;
    }
}

/*----------------------------------------------------------------------*
 |    EOF (%M%)
 *----------------------------------------------------------------------*/
